from . import csv_loader
from . import xlsx_loader
from . import yaml_loader

data_plugins = {
    "csv": csv_loader.load,
    "xlsx": xlsx_loader.load,
    "yaml": yaml_loader.load,
    "yml": yaml_loader.load
}


import sys
sys.path.insert(0,'../')
import pprint

from ttr import ttr

def test_jinja2_render_csv_1():
    generator = ttr("./mock_data/csv_data_1.csv")
    generator.run()
    # pprint.pprint(generator.results)
    assert generator.results == {'r1': 'hostname r1\n!\ninterface loopback0\n  ip address 1.1.1.1/32',
               'r2': '',
               'r3': ''}
               
# test_jinja2_render_csv_1()


"""
Jinja2 Renderer Plugin
**********************

**Prerequisites:** 

- Requires `jinja2 <https://pypi.org/project/Jinja2/>`_ library

This renderer uses Jinja2 templates to render data and produce
text results.

For example, if this is a data to render expressed in YAML format::

    - interface: Gi1/1
      description: Customer A
      dot1q: 100
      ip: 10.0.0.1
      mask: 255.255.255.0
      vrf: cust_a
      template: ttr://interfaces.cisco_ios
      device: rt-1
    - interface: Gi1/2
      description: Customer B
      dot1q: 200
      ip: 10.0.2.1
      mask: 255.255.255.0
      vrf: cust_b
      template: ttr://interfaces.cisco_ios
      device: rt-2
      
``template_name_key`` corresponds to ``template`` key in above data, 
``result_name_key`` corresponds to ``device`` key in above data.

And this is the content of ``ttr://interfaces.cisco_ios`` template::

    interface {{ interface }}
     description {{ description }}
     encapsulation dot1q {{ vid }}
     vrf forwarding  {{ vrf }}
     ip address {{ ip }} {{ mask }}
     ipv6 address {{ ipv6 }}/{{ maskv6 }}
    !
    
This renderer will combine each item in above data with 
``ttr://interfaces.cisco_ios`` template and return results for further 
processing.
"""
import logging
import jinja2
import os
from ..templates import templates_loaders_plugins

log = logging.getLogger(__name__)


def render(
        data, 
        template_name_key,
        templates_dir, 
        templates_dict, 
        result_name_key,
        **renderer_kwargs
    ):
    """
    Render function takes data, templates and produces text output.
    
    :param data: (list), list of dictionaries render
    :param templates_dict: (dict), dictionary keyed by template name with template content as a value
    :param template_name_key: (str), name of template key to use for data rendering, default - ``template``
    :param result_name_key: (str), name of result key to use to combine rendering results, default - ``device``
    :param renderer_kwargs: (dict), kwargs to pass on to ``jinja2.Template(.., **kwargs)`` object instantiation
        
    By default ``renderer_kwargs`` will include::
    
        {"trim_blocks": True, "lstrip_blocks": True}
    """
    result = {}
    templates_objects_dict = {}
    
    renderer_kwargs.setdefault("trim_blocks", True)
    renderer_kwargs.setdefault("lstrip_blocks", True)
    
    # iterate over data and render it
    for datum in data:
        result_name = datum[result_name_key]
        result.setdefault(result_name, [])
        template_name = datum[template_name_key]
        
        # get template content
        if templates_loaders_plugins["base"](template_name, templates_dict, templates_dir):
            template_content = templates_dict[template_name]
        else:
            log.error(
                "TTR:Jinja2 renderer failed to load template file: '{}'".format(
                    template_name
                )
            )    
            continue

        # create Jinja2 template object
        if template_name not in templates_objects_dict:
            try:
                templates_objects_dict[template_name] = jinja2.Template(
                    template_content,
                    **renderer_kwargs
                )
            except Exception as e:
                log.error(
                    "Jinja2 renderer failed to load template: {}; error: {}".format(
                        template_name, e
                    )
                )    
                continue
        
        # render data
        try:
            result[result_name].append(
                templates_objects_dict[template_name].render(datum)
            )
        except Exception as e:
            log.error(
                "Jinja2 renderer failed to render template: {}; error: {}; data: {}".format(
                    template_name, e, datum
                )
            )    
    # transform results into strings
    for result_name, result_list in result.items():
        result[result_name] = "\n".join(result_list)
        del(result_list)
        
    return result



import sys
sys.path.insert(0,'../')
import pprint

from ttr import ttr


def test_csv_data_plugin_load_from_file():
    generator = ttr("./mock_data/csv_data_1.csv")
    assert generator.data_loaded == [
        {'device': 'r1', 'hostname': 'r1', 'lo0_ip': '1.1.1.1', 'template': 'foo'},
        {'device': 'r2', 'hostname': 'r2', 'lo0_ip': '2.2.2.2', 'template': 'bar'},
        {'device': 'r3', 'hostname': 'r3', 'lo0_ip': '3.3.3.3', 'template': 'foobar'}
    ]
    
# test_csv_data_plugin_load_from_file()


def test_openpyxl_data_plugin():
    generator = ttr("./mock_data/table_data_1.xlsx")
    # pprint.pprint(generator.data_loaded)
    assert generator.data_loaded == [{'device': 'r1', 'hostname': 'r1', 'lo0_ip': '1.1.1.1', 'template': 'foo'},
                                     {'device': 'r2', 'hostname': 'r2', 'lo0_ip': '2.2.2.2', 'template': 'bar'},
                                     {'device': 'r3', 'hostname': 'r3', 'lo0_ip': '3.3.3.3', 'template': 'foobar'}]
    
# test_openpyxl_data_plugin()


def test_openpyxl_data_plugin_multitab_table():
    generator = ttr("./mock_data/table_multitab_data_2.xlsx")
    # pprint.pprint(generator.data_loaded)
    assert generator.data_loaded == [{'device': 'r1',
                'hostname': 'r1',
                'lo0_ip': '1.1.1.1',
                'template': 'test_path/device_base'},
               {'device': 'r2',
                'hostname': 'r2',
                'lo0_ip': '2.2.2.2',
                'template': 'test_path/device_base'},
               {'device': 'r1',
                'interface': '10.0.0.1/24',
                'ip': 'Eth1',
                'template': 'test_path/interf_cfg'},
               {'device': 'r1',
                'interface': '10.0.1.1/24',
                'ip': 'Eth2',
                'template': 'test_path/interf_cfg'}]
    generator.run()
    assert generator.results == {'r1': 'hostname r1\n!\ninterface loopback0\n  ip address 1.1.1.1/32\ninterface 10.0.0.1/24\n  ip address Eth1/32\ninterface 10.0.1.1/24\n  ip address Eth2/32', 'r2': 'hostname r2\n!\ninterface loopback0\n  ip address 2.2.2.2/32'}
    
# test_openpyxl_data_plugin_multitab_table()


def test_openpyxl_data_plugin_templates_in_table():
    generator = ttr("./mock_data/table_multitab_inline_templates_data_3.xlsx")
    # pprint.pprint(generator.templates_dict)
    assert generator.templates_dict == {'test_path/device_base': 'hostname {{ hostname }}\n'
                                                                 '!\n'
                                                                 'interface loopback0\n'
                                                                 '  ip address {{ lo0_ip }} 255.255.255.255',
                                        'test_path/interf_cfg': 'interface {{ interface }}\n'
                                                                '  ip address {{ ip }}'}
    # pprint.pprint(generator.data_loaded)                     
    assert generator.data_loaded == [{'device': 'r1',
                                      'hostname': 'r1',
                                      'lo0_ip': '1.1.1.1',
                                      'template': 'test_path/device_base'},
                                     {'device': 'r2',
                                      'hostname': 'r2',
                                      'lo0_ip': '2.2.2.2',
                                      'template': 'test_path/device_base'},
                                     {'device': 'r1',
                                      'interface': 'Eth1',
                                      'ip': '10.0.0.1/24',
                                      'template': 'test_path/interf_cfg'},
                                     {'device': 'r1',
                                      'interface': 'Eth2',
                                      'ip': '10.0.1.1/24',
                                      'template': 'test_path/interf_cfg'}]
    # pprint.pprint(generator.results)
    generator.run()
    assert generator.results == {'r1': 'hostname r1\n'
                                      '!\n'
                                      'interface loopback0\n'
                                      '  ip address 1.1.1.1 255.255.255.255\n'
                                      'interface Eth1\n'
                                      '  ip address 10.0.0.1/24\n'
                                      'interface Eth2\n'
                                      '  ip address 10.0.1.1/24',
                                'r2': 'hostname r2\n'
                                      '!\n'
                                      'interface loopback0\n'
                                      '  ip address 2.2.2.2 255.255.255.255'}
    
test_openpyxl_data_plugin_templates_in_table()

def test_openpyxl_data_plugin_with_multitemplate_processor():
    generator = ttr("./mock_data/table_multiple_templates.xlsx", processors=["multitemplate"])
    # pprint.pprint(generator.data_loaded)
    assert generator.data_loaded == [{'device': 'r1',
                                      'hostname': 'r1',
                                      'lo0_ip': '1.1.1.1',
                                      'template': 'test_path/device_base'},
                                     {'device': 'r1',
                                      'hostname': 'r1',
                                      'lo0_ip': '1.1.1.11',
                                      'template': 'test_path/device_base_rollback'},
                                     {'device': 'r2',
                                      'hostname': 'r2',
                                      'lo0_ip': '2.2.2.2',
                                      'template': 'test_path/device_base'},
                                     {'device': 'r2',
                                      'hostname': 'r2',
                                      'lo0_ip': '2.2.2.22',
                                      'template': 'test_path/device_base_rollback'},
                                     {'device': 'r1',
                                      'interface': 'Eth1',
                                      'ip': '10.0.0.1',
                                      'mask': 24,
                                      'template': 'test_path/interf_cfg'},
                                     {'device': 'r2',
                                      'interface': 'Eth1',
                                      'ip': '10.0.0.2',
                                      'mask': 24,
                                      'template': 'test_path/interf_cfg_b'},
                                     {'device': 'r1',
                                      'interface': 'Eth2',
                                      'ip': '10.0.1.1',
                                      'mask': 24,
                                      'template': 'test_path/interf_cfg'},
                                     {'device': 'r2',
                                      'interface': 'Eth2',
                                      'ip': '10.0.1.2',
                                      'mask': 24,
                                      'template': 'test_path/interf_cfg_b'}]
    # pprint.pprint(generator.templates_dict)
    assert generator.templates_dict == {'test_path/device_base': 'hostname {{ hostname }}\n'
                                                                 '!\n'
                                                                 'interface loopback0\n'
                                                                 '  ip address {{ lo0_ip }} 255.255.255.255',
                                        'test_path/device_base_rollback': 'no hostname {{ hostname }}\n'
                                                                          '!\n'
                                                                          'no interface loopback0',
                                        'test_path/interf_cfg': 'interface {{ interface }}\n  ip address {{ ip }}',
                                        'test_path/interf_cfg_b': 'interface {{ interface }}\n'
                                                                  '  ip address {{ ip }} {{ mask }}'}
    generator.run()
    # pprint.pprint(generator.results)
    assert generator.results == {'r1': 'hostname r1\n'
                                       '!\n'
                                       'interface loopback0\n'
                                       '  ip address 1.1.1.1 255.255.255.255\n'
                                       'no hostname r1\n'
                                       '!\n'
                                       'no interface loopback0\n'
                                       'interface Eth1\n'
                                       '  ip address 10.0.0.1\n'
                                       'interface Eth2\n'
                                       '  ip address 10.0.1.1',
                                 'r2': 'hostname r2\n'
                                       '!\n'
                                       'interface loopback0\n'
                                       '  ip address 2.2.2.2 255.255.255.255\n'
                                       'no hostname r2\n'
                                       '!\n'
                                       'no interface loopback0\n'
                                       'interface Eth1\n'
                                       '  ip address 10.0.0.2 24\n'
                                       'interface Eth2\n'
                                       '  ip address 10.0.1.2 24'}
    
# test_openpyxl_data_plugin_with_multitemplate_processor()

def test_yaml_data_plugin_load_from_text():
    data = """
- interface: Gi1/1
  description: Customer A
  dot1q: 100
  ip: 10.0.0.1
  mask: 255.255.255.0
  vrf: cust_a
- interface: Gi1/2
  description: Customer B
  dot1q: 200
  ip: 10.0.2.1
  mask: 255.255.255.0
  vrf: cust_b
    """
    generator = ttr()
    generator.load_data(data, data_plugin="yaml")
    # pprint.pprint(generator.data_loaded)
    assert generator.data_loaded == [{'description': 'Customer A',
                                      'dot1q': 100,
                                      'interface': 'Gi1/1',
                                      'ip': '10.0.0.1',
                                      'mask': '255.255.255.0',
                                      'vrf': 'cust_a'},
                                     {'description': 'Customer B',
                                      'dot1q': 200,
                                      'interface': 'Gi1/2',
                                      'ip': '10.0.2.1',
                                      'mask': '255.255.255.0',
                                      'vrf': 'cust_b'}]
                                      
def test_yaml_data_plugin_load_from_file():
    generator = ttr("./mock_data/yaml_data_1.yaml")
    assert generator.data_loaded == [{'description': 'Customer A',
                                      'dot1q': 100,
                                      'interface': 'Gi1/1',
                                      'ip': '10.0.0.1',
                                      'mask': '255.255.255.0',
                                      'vrf': 'cust_a'},
                                     {'description': 'Customer B',
                                      'dot1q': 200,
                                      'interface': 'Gi1/2',
                                      'ip': '10.0.2.1',
                                      'mask': '255.255.255.0',
                                      'vrf': 'cust_b'}]
# test_yaml_data_plugin_load_from_file()


import sys
sys.path.insert(0,'../')
import pprint
import shutil
import os

from ttr import ttr

def test_file_returner_1():
    # delete previously generated results
    if os.path.exists("./Output/test_file_returner_1"):
        shutil.rmtree("./Output/test_file_returner_1")
    generator = ttr(
        "./mock_data/csv_data_1.csv", 
        returner="file", 
        returner_kwargs={
            "result_dir": "./Output/test_file_returner_1/"
        }
    )
    generator.run()
    with open("./Output/test_file_returner_1/r1.txt") as f:
        assert f.read() == """hostname r1
!
interface loopback0
  ip address 1.1.1.1/32"""
    with open("./Output/test_file_returner_1/r2.txt") as f:
        assert f.read() == ""
    with open("./Output/test_file_returner_1/r3.txt") as f:
        assert f.read() == ""   
        
def test_terminal_returner_1():
    generator = ttr(
        "./mock_data/csv_data_1.csv", 
        returner="terminal"
    )

# test_terminal_returner_1()



"""
TTR CLI tool
############

This tool can be used to render data from various sources and either
print results to screen or save them in output folder.

Supported arguments::

    -d,  --data          OS path to folder with data files or to data file, default ./Data/
    -t,  --templates     OS path to folder with templates, default ./Templates/
    -o,  --output        Output folder location, default ./Output/<current time><data file name>/
    -p,  --print         Print results to terminal instead of saving to folder
    -l,  --logging       Set logging level - "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"; default ERROR

How to use
**********

In general case TTR CLI utility takes data file and templates location references and saves produced results in a subfolder within ./Output/ directory, where subfolder name has this format ./Output/<current time><data file name>/.

Sample invocation::

    ttr -d ./data/data.yaml

Alternatively a path to directory can be provided instead of data file, in that case TTR will scan that path and prompt
user to select file to work with::

    ttr -d ./data/
    ====================
    Files found in './mock_data/' directory
    0: csv_data_1.csv
    1: data.yaml
    2: table_data_1.xlsx
    Choose data file to work with (number): 1


"""
import argparse
import time
import os
import logging

log = logging.getLogger(__name__)


#if run as a script, inject ttr folder in system path
if __name__ == "__main__":
    import sys
    sys.path.insert(0, '.')

from ttr import ttr

__version__ = "0.1.0"
ctime = time.strftime("%Y-%m-%d_%H-%M-%S")

cli_help = """
Template Text Renderer CLI utility

-d,  --data          OS path to folder with data files or to data file, default ./Data/
-t,  --templates     OS path to folder with templates, default ./Templates/
-o,  --output        Output folder location, default ./Output/<current time><data file name>/
-p,  --print         Print results to terminal instead of saving to folder
-l,  --logging       Set logging level - "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"; default ERROR
"""

def cli_tool():
    # form argparser menu:
    description_text = """{}""".format(cli_help)

    argparser = argparse.ArgumentParser(
        description="TTR CLI, version {}".format(__version__),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    run_options = argparser.add_argument_group(description=description_text)
    #-----------------------------------------------------------------------------
    # General options
    #-----------------------------------------------------------------------------
    run_options.add_argument(
        "-d",
        "--data",
        action="store",
        dest="DATA",
        default="./Data/",
        type=str,
        help=argparse.SUPPRESS,
    )
    run_options.add_argument(
        "-t",
        "--templates",
        action="store",
        dest="TEMPLATES_FOLDER",
        default="./Templates/",
        type=str,
        help=argparse.SUPPRESS,
    )
    run_options.add_argument(
        "-o",
        "--output",
        action="store",
        dest="OUTPUT_FOLDER",
        default='./Output/{}_{{}}/'.format(ctime),
        type=str,
        help=argparse.SUPPRESS,
    )
    run_options.add_argument(
        "-p",
        "--print",
        action="store_true",
        dest="PRINT_TO_TERMINAL",
        default=False,
        help=argparse.SUPPRESS,
    )
    run_options.add_argument(
        "-l",
        "--logging",
        action="store",
        dest="LOGGING_LEVEL",
        default='WARNING',
        type=str,
        help=argparse.SUPPRESS,
    )
    #-----------------------------------------------------------------------------
    # Parse arguments
    #-----------------------------------------------------------------------------
    args = argparser.parse_args()

    # general arguments
    DATA = args.DATA  # string, OS path to data file or folder with files to process
    TEMPLATES_FOLDER = args.TEMPLATES_FOLDER  # OS path to folder to save results into
    OUTPUT_FOLDER = args.OUTPUT_FOLDER  # output filename
    PRINT_TO_TERMINAL = args.PRINT_TO_TERMINAL
    LOGGING_LEVEL = args.LOGGING_LEVEL

    # set logging level
    try:
        logging.basicConfig(
            format="%(asctime)s.%(msecs)d [TTR %(levelname)s] %(lineno)d; %(message)s",
            datefmt="%m/%d/%Y %I:%M:%S",
            level=getattr(logging, LOGGING_LEVEL.upper())
        )
    except Exception as e:
        log.error("TTR:cli_tool - Failed to set logging level to '{}', error: {}".format(LOGGING_LEVEL, e))

    # get data file OS path
    if os.path.isdir(DATA):
        # get the list of files in 'Data' folder
        data_files = os.listdir(DATA)
        if len(data_files) == 1:
            file_number = 0
        elif len(data_files) == 0:
            log.error("No files found in 'Data' directory; Exiting...")
            raise SystemExit()
        # if more than one file found, ask user which file to use
        else:
            print("{}\nFiles found in '{}' directory".format(20 * "=", DATA))
            for position, item in enumerate(data_files):
                print("{}: {}".format(position, item))
            # get file number to work with
            file_number = -1
            while not(0 <= file_number <= len(data_files) - 1):
                try:
                    file_number = int(input("Choose data file to work with (number): "))
                except KeyboardInterrupt:
                    raise SystemExit()
                except:
                    continue
        # form path to data file
        data_file_path = os.path.join(DATA, data_files[file_number])
    elif os.path.isfile(DATA):
        data_file_path = DATA
    else:
        log.error("Failed to form data_file_path, data: '{}'".format(DATA))
        return

    if not PRINT_TO_TERMINAL and not os.path.exists(OUTPUT_FOLDER):
        OUTPUT_FOLDER = OUTPUT_FOLDER.format(
            os.path.split(data_file_path)[-1]
        )
        os.makedirs(OUTPUT_FOLDER)

    # generate results and save them in output folder or print to screen
    with ttr(
            data=data_file_path,
            templates_dir=TEMPLATES_FOLDER,
            returner="terminal" if PRINT_TO_TERMINAL else "file",
            returner_kwargs={
                "result_dir": OUTPUT_FOLDER
            },
            processors = ["multitemplate"]
    ) as g:
        g.run()

if __name__ == "__main__":
    cli_tool()



"""
Terminal Returner Plugin
************************

**Plugin Name:** ``terminal``

This plugin prints rendered result to terminal screen
applying minimal formating to improve readability.

For instance if these are rendering results::

    {'rt-1': 'interface Gi1/1\\n'
             ' description Customer A\\n'
             ' encapsulation dot1q 100\\n'
             ' vrf forwarding  cust_a\\n'
             ' ip address 10.0.0.1 255.255.255.0\\n'
             ' exit\\n'
             '!\\n'
             'interface Gi1/2\\n'
             ' description Customer C\\n'
             ' encapsulation dot1q 300\\n'
             ' vrf forwarding  cust_c\\n'
             ' ip address 10.0.3.1 255.255.255.0\\n'
             ' exit\\n'
             '!',
     'rt-2': 'interface Gi1/2\\n'
             ' description Customer B\\n'
             ' encapsulation dot1q 200\\n'
             ' vrf forwarding  cust_b\\n'
             ' ip address 10.0.2.1 255.255.255.0\\n'
             ' exit\\n'
             '!'}
					
Terminal returner will print to screen::

    # ---------------------------------------------------------------------------
    # rt-1 rendering results
    # ---------------------------------------------------------------------------
    interface Gi1/1
     description Customer A
     encapsulation dot1q 100
     vrf forwarding  cust_a
     ip address 10.0.0.1 255.255.255.0
     exit
    !
    interface Gi1/2
     description Customer C
     encapsulation dot1q 300
     vrf forwarding  cust_c
     ip address 10.0.3.1 255.255.255.0
     exit
    !
    
    # ---------------------------------------------------------------------------
    # rt-2 rendering results
    # ---------------------------------------------------------------------------
    interface Gi1/2
     description Customer B
     encapsulation dot1q 200
     vrf forwarding  cust_b
     ip address 10.0.2.1 255.255.255.0
     exit
    !
	
This returner useful for debugging or, for instance, when it is easier 
to copy produced results from terminal screen.
"""

def dump(data_dict, **kwargs):
    """
    This function prints results to terminal screen
    
    :param data_dict: (dict) dictionary keyed by ``result_name_key`` where 
                      values are rendered results string
    """
    for key, value in data_dict.items():
        print("""
# ---------------------------------------------------------------------------
# {} rendering results
# ---------------------------------------------------------------------------""".format(key))
        print(value)


"""
Self Returner Plugin
********************

**Plugin Name:** ``self``

This plugin does nothing with results, implementing
behaviour where results stored in TTR object for 
further programmatic consumption.
"""

def dump(data_dict, **kwargs):
    """
    This function applying no actions to results,
    and implemented to keep plugins API consistent.
    
    :param data_dict: (dict) dictionary keyed by ``result_name_key`` where 
                      values are rendered results string
    """
    pass


from . import file_returner
from . import terminal_returner
from . import self_returner

returners_plugins = {
    "self": self_returner.dump,
    "file": file_returner.dump,
    "terminal": terminal_returner.dump
}


"""
Multitemplate Processor
***********************

**Plugin reference name:** ``multitemplate``

Processor to extract multiple template dictionaries from each 
data item based on suffix values.

Takes a list of dictionaries::

    [{'device': 'r1',
      'hostname': 'r1',
      'lo0_ip': '1.1.1.1',
      'lo0_ip_rollback': '1.1.1.11',
      'template': 'test_path/device_base',
      'template_rollback': 'test_path/device_base_rollback'},
     {'device:a': 'r1',
      'device:b': 'r2',
      'interface:a': 'Eth1',
      'interface:b': 'Eth1',
      'ip:a': '10.0.0.1',
      'ip:b': '10.0.0.2',
      'mask': 24,
      'template:a': 'test_path/interf_cfg',
      'template:b': 'test_path/interf_cfg_b'}]

Returns::

    [{'device': 'r1',
    'hostname': 'r1',
    'lo0_ip': '1.1.1.11',
    'template': 'test_path/device_base_rollback'},
    {'device': 'r1',
    'hostname': 'r1',
    'lo0_ip': '1.1.1.1',
    'template': 'test_path/device_base'},
    {'device': 'r1',
    'interface': 'Eth1',
    'ip': '10.0.0.1',
    'mask': 24,
    'template': 'test_path/interf_cfg'},
    {'device': 'r2',
    'interface': 'Eth1',
    'ip': '10.0.0.2',
    'mask': 24,
    'template': 'test_path/interf_cfg_b'}]

Where ``template_name_key`` is ``template``.

Multitemplate processor detects on endings, ``:a`` and ``:b``
for instance and uses them to split dictionaries in several dictionaries,
populated with values corresponding to certain suffixes.

Key names without suffixes considered as common values and included in
all dictionaries produced.
"""
import logging
import pprint

log = logging.getLogger(__name__)


def process(data, template_name_key, **kwargs):
    """
    Function to process multitemplate data items.
    
    :param data: (list), data to process - list of dictionaries
    :param template_name_key: string, name of the template key
    """
    ret = []
    headers = tuple()
    previous_headers = tuple()
    headers_to_endings = {}

    # scan through data to split in multiple items
    for datum in data:

        headers = tuple(datum.keys())

        # check if need to form headers_to_endings dictionary of {ending: [headers]}
        if headers != previous_headers:
            endings = tuple(
                [
                    h.replace(template_name_key, "")
                    for h in headers
                    if h.startswith(template_name_key) and h != template_name_key
                ]
            )
            headers_without_endings = [h for h in headers if not h.endswith(endings)]
            headers_to_endings = {
                ending: headers_without_endings
                + [h for h in headers if h.endswith(ending)]
                for ending in endings
            }
            if template_name_key in headers_without_endings:
                headers_to_endings = {"": headers_without_endings, **headers_to_endings}
        # form data
        for ending, headers_item in headers_to_endings.items():
            ret.append(
                {
                    h[: -len(ending)] if h.endswith(ending) and ending else h: datum[h]
                    for h in headers_item
                }
            )
        previous_headers = headers
    return ret



"""
CSV Spreadsheets loader plugin
******************************

**Plugin Name:** ``csv``

Support loading data from CSV text file.

Spreadsheet must contain a column or multiple columns with headers starting 
with ``template_name_key`` argument string. Values of template(s) columns either 
names of the template to use for rendering or OS path string to template file 
relative to ``template_dir`` argument supplied to TTR object on instantiation.

In addition, table must contain column with ``result_name_key`` values, they used
to combine results, i.e. rendering results for identical ``result_name_key`` combined
in a single string. ``result_name_key`` used further by returners to return results.
"""
import logging
import csv
import os

log = logging.getLogger(__name__)

def load(data, templates_dict, template_name_key, **kwargs):
    """
    Function to load CSV spreadsheet.
    
    :param data: OS path to CSV text file
    :param templates_dict: dictionary to load templates from spreadsheet, not supported by csv loader
    :param template_name_key: string, templates column header prefix
    """
    ret = None
    
    # load from file
    if os.path.isfile(data):
        with open(data, newline='') as csvfile:
            reader = csv.DictReader(csvfile, **kwargs)
            ret = [dict(row) for row in reader]
    # load all csv files from folder
    elif os.path.isdir(data):
        pass
    # load data text as is using stringio module
    elif isinstance(data, str):
        pass
    else:
        raise SystemExit("csv_loader, unsupported data, should be either OS path to file, directory or text")
        
    return ret


"""
File Returner Plugin
********************

**Plugin Name:** ``file``

This plugin responsible for saving results to text files
iterating over results dictionary keyed by ``result_name_key``.

For example, if results ``data_dict`` might look like this::

    {"rt-1": "interface Gi1/1\\n"
             " description Customer A\\n"
             " encapsulation dot1q 100\\n"
             " vrf forwarding  cust_a\\n"
             " ip address 10.0.0.1 255.255.255.0\\n"
             " exit\\n"
             "!\\n"
             "interface Gi1/2\\n"
             " description Customer C\\n"
             " encapsulation dot1q 300\\n"
             " vrf forwarding  cust_c\\n"
             " ip address 10.0.3.1 255.255.255.0\\n"
             " exit\\n"
             "!",
     "rt-2": "interface Gi1/2\\n"
             " description Customer B\\n"
             " encapsulation dot1q 200\\n"
             " vrf forwarding  cust_b\\n"
             " ip address 10.0.2.1 255.255.255.0\\n"
             " exit\\n"
             "!"}

If ``result_dir`` argument set to ``./Output/``, file returner
will iterate over ``data_dict`` using keys as filenames
populating files with values at the end ``./Output/`` directory
will contain two files named ``rt-1.txt`` and ``rt-2.txt`` with 
respective content.    
"""
import logging
import os


def dump(
        data_dict, 
        result_dir="./Output/",
        **kwargs
    ):
    """
    Function to save results in text files.
    
    :param data_dict: (dict) dictionary keyed by ``result_name_key`` where 
                      values are strings to save in text files
    :param result_dir: (str) OS path to directory to save results in
    """
    if not os.path.exists(result_dir):
        os.makedirs(result_dir)
    
    for datum_name, filedata in data_dict.items():
        filename = os.path.join(
            result_dir, "{}.txt".format(
                datum_name
            )
        )
        with open(filename, "w", encoding="utf-8") as f:
            f.write(filedata)



"""
Template Text Renderer - TTR
############################

Module to produce text files using templates. TTR targets to implement common work flow:

.. image:: ./_images/workflow.png

Above approach is fairly simple but powerful enough to address various use cases
where structured data need to be transformed in a textual form understandable by
targeted system.

TTR uses plugins to load data and templates, render and return results.

.. image:: ./_images/Plugins_overview.png

In addition, TTR comes with a collection of Jinja2 templates to help with
common use cases.
"""
__version__ = "0.1.0"

import logging
import os
from .plugins.data import data_plugins
from .plugins.renderers import renderers_plugins
from .plugins.returners import returners_plugins
from .plugins.processors import processors_plugins

log = logging.getLogger(__name__)


class ttr:
    """
    Main class to instantiate TTR object.

    :param data: (str) type depends on data plugin in use, but can be an OS
        path string referring to YAML structured text file or CSV spreadsheet
    :param data_plugin: (str) name of data plugin to use to load data
    :param data_plugin_kwargs: (dict) arguments to pass on to data plugin
    :param renderer: (str) name of renderer plugin to use, default ``jinja2``
    :param renderer_kwargs: (dict) arguments to pass on to renderer plugin
    :param templates_dir: (str) OS pat to directory with templates, default ``./Templates/``
    :param template_name_key: (str) name of the key in data items that reference template
        to use to render that particular datum, default ``template``
    :param returner: (str) name of returner plugin to use, default ``self``
    :param returner_kwargs: (dict) arguments to pass on to returner plugin
    :param result_name_key: (str) name of the key in data items value of which should be
        used as a key in results dictionary, default ``device``
    :param processors: (list) list of processor plugins names to pass loaded data
        through, default is empty list - no processors applied
    :param templates_dict: (dict) dictionary of {template_name: template_content}
    """
    def __init__(
        self,
        data=None,
        data_plugin=None,
        data_plugin_kwargs={},
        renderer="jinja2",
        renderer_kwargs={},
        templates_dir="./Templates/",
        template_name_key="template",
        returner="self",
        returner_kwargs={},
        result_name_key="device",
        processors=[],
        templates_dict={}
    ):
        self.data_plugin = data_plugin
        self.data_plugin_kwargs = data_plugin_kwargs
        self.templates_dir = templates_dir
        self.templates_dict = templates_dict or {}
        self.template_name_key = template_name_key
        self.renderer = renderer
        self.renderer_kwargs = renderer_kwargs
        self.returner = returner
        self.returner_kwargs = returner_kwargs
        self.result_name_key = result_name_key
        self.results = None
        self.processors = processors
        self.data_loaded = None

        # load data to render
        if data:
            self.load_data(data)


    def __enter__(self):
        return self


    def __exit__(self, type, value, traceback):
        del self.data_loaded, self.templates_dict, self.results


    def load_data(self, data, data_plugin=None):
        """
        Method to load data to render.

        :param data: (str) data to load, either OS path to data file or text
        :param data_plugin: (str) name of data plugin to load data, by default
            will choose data plugin based on file extension
        """
        if data_plugin:
            plugin_name = data_plugin
        elif self.data_plugin:
            plugin_name = self.data_plugin
        elif os.path.isfile(data[:5000]):
            # get data loader name based on data file extension
            plugin_name = data.split(".")[-1].strip()
        else:
            log.error("load_data: failed to identify data plugin to load data '{}'".format(data[:100]))
            return

        # load data using plugin
        log.debug("Loading data using '{}' plugin".format(plugin_name))
        self.data_plugin_kwargs.setdefault("template_name_key", self.template_name_key)
        self.data_plugin_kwargs["templates_dict"] = self.templates_dict
        self.data_loaded = data_plugins[plugin_name](data, **self.data_plugin_kwargs)

        # process loaded data
        for processor_plugin in self.processors:
            log.debug("Running loaded data through processor: '{}'".format(processor_plugin))
            self.data_loaded = processors_plugins[processor_plugin](
                self.data_loaded,
                data_plugin=self.data_plugin,
                template_name_key=self.template_name_key
            )

        log.debug("Data '{}' loaded".format(str(data)[:40]))

    def run(self):
        """
        Method to render templates with data and produce dictionary results
        keyed by ``result_name_key``.

        If returner set to ``self``, will return results dictionary.
        """
        # generate results
        log.debug("Rendering data using '{}' renderer".format(self.renderer))
        self.results = renderers_plugins[self.renderer](
            self.data_loaded,
            self.template_name_key,
            self.templates_dir,
            self.templates_dict,
            self.result_name_key,
            **self.renderer_kwargs
        )
        # return results
        log.debug("Returning results using '{}' returner".format(self.returner))
        returners_plugins[self.returner](
            self.results,
            **self.returner_kwargs
        )
        log.debug("TTR rendering run completed")
        if self.returner == "self":
            return self.results



"""
XLSX Spreadsheets loader plugin
*******************************

**Plugin Name:** ``xlsx``

This plugin supports loading data from multiple sheets, combining them for rendering.

**Prerequisites:**

- Requires `openpyxl <https://pypi.org/project/openpyxl/>`_ library

Spreadsheets must contain a column or multiple columns with headers starting
with ``template_name_key`` argument string. Values of template(s) columns either
names of the template to use for rendering or OS path string to template file
relative to ``template_dir`` argument supplied to TTR object on instantiation.

In addition, table must contain column with ``result_name_key`` values, they used
to combine results, i.e. rendering results for identical ``result_name_key`` combined
in a single string. ``result_name_key`` used further by returners to return results.

Spreadsheet might contain multiple tabs with names starting with ``template``, these
tabs can contain templates to use for rendering data from other tabs. All the templates
loaded line by line, ``template:{{ template_name }}`` lines used to identify end
of previous and start of next template, where ``template_name`` used for referencing
template in templates columns.
"""

import logging
from openpyxl import load_workbook

log = logging.getLogger(__name__)


def load_template(sheet, templates_dict):
    log.debug("XLSX loader, loading templates tab - '{}'".format(sheet.title))
    current_template_name = ""
    current_template_lines = []
    for item in sheet.iter_rows(min_row=1, max_col=1, values_only=True):
        rowData = item[0]
        if rowData != None:  # check if cell is empty and skip it if so
            if rowData.upper().startswith("TEMPLATE:"):
                if current_template_lines and current_template_name:
                    templates_dict[current_template_name] = "\n".join(
                        current_template_lines
                    )
                current_template_name = rowData.split(":")[1].strip()
                current_template_lines = []
            else:
                current_template_lines.append(rowData)
    # add last template
    if current_template_lines and current_template_name:
        templates_dict[current_template_name] = "\n".join(current_template_lines)


def load_data_from_sheet(sheet, ret, template_name_key):
    headers = [
        sheet.cell(row=1, column=i).value for i in range(1, sheet.max_column + 1)
        if not sheet.cell(row=1, column=i).value.startswith("#")
    ]

    has_templates_column = False
    for header in headers:
        if header.startswith(template_name_key):
            has_templates_column = True
            break

    if not has_templates_column:
        log.warning(
            "XLSX loader, no '{}' header on tab '{}', skipping it".format(
                template_name_key, sheet.title
            )
        )
        return

    # form data
    log.debug("XLSX loader, loading data - tab: '{}', headers: '{}'".format(sheet.title, headers))
    for row in sheet.iter_rows(min_row=2, values_only=True):
        ret.append(
            dict(zip(headers, row))
        )


def load(data, templates_dict, template_name_key, **kwargs):
    """
    Function to load XLSX spreadsheet. Takes OS path to ``.xlsx`` file
    and returns list of dictionaries, where keys equal to headers
    of spreadsheets' tabs.

    :param data: string, OS path to ``.xlsx`` file
    :param templates_dict: dictionary to load templates from spreadsheet
    :param template_name_key: string, templates column header prefix
    """
    ret = []

    kwargs.setdefault("data_only", True)
    kwargs.setdefault("read_only", True)

    wb = load_workbook(data, **kwargs)

    for sheet_name in wb.sheetnames:
        if sheet_name.startswith("#"):
            log.debug("XLSX loader, skipping tab - '{}'".format(sheet_name))
            continue
        elif "TEMPLATE" in sheet_name.upper():
            load_template(wb[sheet_name], templates_dict)
        else:
            load_data_from_sheet(wb[sheet_name], ret, template_name_key)
    return ret



import sys
sys.path.insert(0,'../')
import pprint

from ttr import ttr


def test_loading_template_from_collection():
    data = """
- interface: Gi1/1
  description: Customer A
  vid: 100
  ip: 10.0.0.1
  mask: 255.255.255.0
  vrf: cust_a
  template: ttr://interfaces.cisco_ios.txt
  device: rt-1
- interface: Gi1/2
  description: Customer C
  vid: 300
  ip: 10.0.3.1
  mask: 255.255.255.0
  vrf: cust_c
  template: ttr://interfaces.cisco_ios.txt
  device: rt-1
- interface: Gi1/2
  description: Customer B
  vid: 200
  ip: 10.0.2.1
  mask: 255.255.255.0
  vrf: cust_b
  template: ttr://interfaces.cisco_ios.txt
  device: rt-2
    """
    generator = ttr(data=data, data_plugin="yaml")
    generator.run()
    # pprint.pprint(generator.results)
    assert generator.results == {'rt-1': 'interface Gi1/1\n'
                                         ' description Customer A\n'
                                         ' encapsulation dot1q 100\n'
                                         ' vrf forwarding  cust_a\n'
                                         ' ip address 10.0.0.1 255.255.255.0\n'
                                         ' exit\n'
                                         '!\n'
                                         'interface Gi1/2\n'
                                         ' description Customer C\n'
                                         ' encapsulation dot1q 300\n'
                                         ' vrf forwarding  cust_c\n'
                                         ' ip address 10.0.3.1 255.255.255.0\n'
                                         ' exit\n'
                                         '!',
                                 'rt-2': 'interface Gi1/2\n'
                                         ' description Customer B\n'
                                         ' encapsulation dot1q 200\n'
                                         ' vrf forwarding  cust_b\n'
                                         ' ip address 10.0.2.1 255.255.255.0\n'
                                         ' exit\n'
                                         '!'}
         
# test_loading_template_from_collection()

def test_loading_template_from_collection_no_txt_extension():
    data = """
- interface: Gi1/1
  description: Customer A
  vid: 100
  ip: 10.0.0.1
  mask: 255.255.255.0
  vrf: cust_a
  template: ttr://interfaces.cisco_ios
  device: rt-1
- interface: Gi1/2
  description: Customer B
  vid: 200
  ip: 10.0.2.1
  mask: 255.255.255.0
  vrf: cust_b
  template: ttr://interfaces.cisco_ios
  device: rt-2
    """
    generator = ttr(data=data, data_plugin="yaml")
    generator.run()
    # pprint.pprint(generator.results)
    assert generator.results == {'rt-1': 'interface Gi1/1\n'
                                         ' description Customer A\n'
                                         ' encapsulation dot1q 100\n'
                                         ' vrf forwarding  cust_a\n'
                                         ' ip address 10.0.0.1 255.255.255.0\n'
                                         ' exit\n'
                                         '!',
                                 'rt-2': 'interface Gi1/2\n'
                                         ' description Customer B\n'
                                         ' encapsulation dot1q 200\n'
                                         ' vrf forwarding  cust_b\n'
                                         ' ip address 10.0.2.1 255.255.255.0\n'
                                         ' exit\n'
                                         '!'}
         
# test_loading_template_from_collection_no_txt_extension()


import sys
sys.path.insert(0,'../')
import pprint

from ttr import ttr

def test_context_manager_1():
    with ttr("./mock_data/csv_data_1.csv") as g:
        assert g.data_loaded == [
            {'device': 'r1', 'hostname': 'r1', 'lo0_ip': '1.1.1.1', 'template': 'foo'},
            {'device': 'r2', 'hostname': 'r2', 'lo0_ip': '2.2.2.2', 'template': 'bar'},
            {'device': 'r3', 'hostname': 'r3', 'lo0_ip': '3.3.3.3', 'template': 'foobar'}
        ]
        results = g.run()
        assert results == {'r1': 'hostname r1\n!\ninterface loopback0\n  ip address 1.1.1.1/32',
                'r2': '',
                'r3': ''}


"""
Base Template Loader
********************

Base loader loads templates from ``.txt`` files. 

.. warning:: Only ``.txt`` files supported by this loader unless ``template_name`` referring existing file.

This loader searches for templates in this order:

0. Check if template with given name already loaded, use it if so
1. Check if template name starts with ``ttr://``, load it from TTR package if so
2. If template name references file, load it
3. Use ``templates_dir`` as a directory to search for template file.

For cases 1 and 3 template name can omit extension, but base loader will assume
``.txt`` extension in that case. 

On failure to load template file, base loader will log an error message and
TTR continue processing of other data items.
"""

import os
import logging

log = logging.getLogger(__name__)


def load(template_name, templates_dict, templates_dir):
    """
    Function to locate template file and return it's content
    
    **Attributes**
    
    * template_name (str) - OS path to template file relative to execution folder 
      or to /ttr/templates/ package directory or name of template within spreadsheet
    * templates_dir (str) - name of directory with templates
    * templates_dict (dict) - dictionary of templates content
    
    **Valid combinations of template location**
       
    * ``template_name="./misc/foo/bar.txt"`` 
    * ``template_name="ttr://misc/foo/bar.txt"``   
    
    On success loads template content in ``templates_dict`` and returns ``True``, on failure
    returns ``False``
    """
    # check if template already loaded
    if template_name in templates_dict:
        return True
    elif not template_name:
        log.warning("TTR:base_template_loader - invalid template_name value: '{}'".format(template_name))
        return False
    # check if template_name referring to template in TTR package
    elif template_name.startswith("ttr://"):
        template_filepath = template_name.replace("ttr://", "")
        if not template_filepath.endswith(".txt"):
            template_filepath = "{}.txt".format(template_filepath)
        # construct path "<abs path>\ttr\plugins\templates\..\..\templates\<template name>" 
        # to move to ttr package dir and down to templates directory
        template_filepath = os.path.join(os.path.abspath(os.path.dirname(__file__)), "..", "..", "templates", template_filepath)
    # check if template_name is a path to file
    elif os.path.isfile(template_name):
        template_filepath = template_name
    # assume that template_name is a path to text file within templates_dir
    else:
        # try to use template_name as is
        template_filepath = os.path.join(templates_dir, template_name)
        # check if path referring a file, if not, append .txt to template file name
        if not os.path.isfile(template_filepath) and not template_name.endswith(".txt"):
            template_filepath = os.path.join(
                templates_dir, "{}.txt".format(template_name)
            )            
            
    # load template content from file
    if os.path.isfile(template_filepath):
        with open(template_filepath, "r") as f:
            templates_dict[template_name] = f.read()        
    else:
        log.error("TTR:base_template_loader - Template not found: {}; formed template_filepath: {}".format(template_name, template_filepath))
        return False

    return True



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

__author__ = "Denis Mulyalin <d.mulyalin@gmail.com>"

setup(
    name="py-ttr",
    version="0.1.0",
    author="Denis Mulyalin",
    author_email="d.mulyalin@gmail.com",
    description="Extendable Bulk Configuration Generator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/dmulyalin/ttr",
    packages=find_packages(),
    package_data={
        "ttr": [
            "templates/*"
        ]
    },
    include_package_data=True,
    data_files=[('', ['LICENSE'])],
    classifiers=[
        "Topic :: Utilities",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points = {
        'console_scripts': ['ttr=ttr:cli_tool'],
    }
)



import sys
sys.path.insert(0,'../')
import pprint

from ttr import ttr

def test_quick_start():
    templates = {"interfaces.cisco_ios.txt": """
interface {{ interface }}
{% if description is defined %}
 description {{ description }}
{% endif %}
{% if vid is defined %}
 encapsulation dot1q {{ vid }}
{% endif %}
{% if vrf is defined %}
 vrf forwarding  {{ vrf }}
{% endif %}
{% if ip is defined and mask is defined %}
 ip address {{ ip }} {{ mask }}
{% endif %}
{% if ipv6 is defined and maskv6 is defined %}
 ipv6 address {{ ipv6 }}/{{ maskv6 }}
{% endif %}
 exit
!
"""
}

    data = """
- interface: Gi1/1
  description: Customer A
  vid: 100
  ip: 10.0.0.1
  mask: 255.255.255.0
  vrf: cust_a
  template: interfaces.cisco_ios.txt
  device: rt-1
- interface: Gi1/2
  description: Customer C
  vid: 300
  ip: 10.0.3.1
  mask: 255.255.255.0
  vrf: cust_c
  template: interfaces.cisco_ios.txt
  device: rt-1
- interface: Gi1/2
  description: Customer B
  vid: 200
  ip: 10.0.2.1
  mask: 255.255.255.0
  vrf: cust_b
  template: interfaces.cisco_ios.txt
  device: rt-2
"""

    gen = ttr(data=data, data_plugin="yaml", templates_dict=templates)
    results = gen.run()    
    # pprint.pprint(results)
    assert results == {'rt-1': '\n'
                               'interface Gi1/1\n'
                               ' description Customer A\n'
                               ' encapsulation dot1q 100\n'
                               ' vrf forwarding  cust_a\n'
                               ' ip address 10.0.0.1 255.255.255.0\n'
                               ' exit\n'
                               '!\n'
                               '\n'
                               'interface Gi1/2\n'
                               ' description Customer C\n'
                               ' encapsulation dot1q 300\n'
                               ' vrf forwarding  cust_c\n'
                               ' ip address 10.0.3.1 255.255.255.0\n'
                               ' exit\n'
                               '!',
                       'rt-2': '\n'
                               'interface Gi1/2\n'
                               ' description Customer B\n'
                               ' encapsulation dot1q 200\n'
                               ' vrf forwarding  cust_b\n'
                               ' ip address 10.0.2.1 255.255.255.0\n'
                               ' exit\n'
                               '!'}

# test_quick_start()

def test_quick_start_from_files():
    gen = ttr(
        data="./mock_data/quick_start_mock_data.yaml", 
        templates_dir="./Templates/"
    )
    gen.run()
    results = gen.results
    # pprint.pprint(results)
    assert results == {'rt-1': 'interface Gi1/1\n'
                               ' description Customer A\n'
                               ' encapsulation dot1q 100\n'
                               ' vrf forwarding  cust_a\n'
                               ' ip address 10.0.0.1 255.255.255.0\n'
                               ' exit\n'
                               '!\n'
                               'interface Gi1/2\n'
                               ' description Customer C\n'
                               ' encapsulation dot1q 300\n'
                               ' vrf forwarding  cust_c\n'
                               ' ip address 10.0.3.1 255.255.255.0\n'
                               ' exit\n'
                               '!',
                       'rt-2': 'interface Gi1/2\n'
                               ' description Customer B\n'
                               ' encapsulation dot1q 200\n'
                               ' vrf forwarding  cust_b\n'
                               ' ip address 10.0.2.1 255.255.255.0\n'
                               ' exit\n'
                               '!'}

# test_quick_start_from_files()

def test_quick_start_from_files_context_manager():
    with ttr("./mock_data/quick_start_mock_data.yaml") as gen:
        results = gen.run()
    # pprint.pprint(results)
    assert results == {'rt-1': 'interface Gi1/1\n'
                               ' description Customer A\n'
                               ' encapsulation dot1q 100\n'
                               ' vrf forwarding  cust_a\n'
                               ' ip address 10.0.0.1 255.255.255.0\n'
                               ' exit\n'
                               '!\n'
                               'interface Gi1/2\n'
                               ' description Customer C\n'
                               ' encapsulation dot1q 300\n'
                               ' vrf forwarding  cust_c\n'
                               ' ip address 10.0.3.1 255.255.255.0\n'
                               ' exit\n'
                               '!',
                       'rt-2': 'interface Gi1/2\n'
                               ' description Customer B\n'
                               ' encapsulation dot1q 200\n'
                               ' vrf forwarding  cust_b\n'
                               ' ip address 10.0.2.1 255.255.255.0\n'
                               ' exit\n'
                               '!'}


"""
YAML loader
***********

**Plugin Name:** ``yaml``

**Prerequisites:** 

- Requires PyYAML library

Plugin to load data to render from YAML structured text.
"""
import logging
import os

log = logging.getLogger(__name__)

try:
    from yaml import safe_load
except ImportError:
    log.error(
        "yaml_loader: failed to import YAML module, install: 'python -m pip install pyyaml'"
    )


def load(data, **kwargs):
    """
    Function to load YAML data from text file or from string. Text file should have
	``.yml`` or ``.yaml`` extension to properly detect loader.
    
    :param data: string, OS path to text file or YAML structured text
    """
    ret = None
    
    # load from file
    if os.path.isfile(data[:1000]):
        with open(data, newline='') as yamlfile:
            ret = safe_load(yamlfile)
    # load as is
    elif isinstance(data, str):
        ret = safe_load(data)
    else:
        raise SystemExit("yaml_loader, unsupported data, should be either OS path to file or text")
        
    return ret